#!/usr/bin/ksh93
################################################################################
#									       #
# Shell script to attempt to repair stale partitions in volume groups.	       #
#									       #
# Please note that the hashbang (first) line in this script causes it to be    #
# run by the enhanced Korn shell (ksh93), which supports associative arrays.   #
#									       #
# Author: Steve Pittman - 925-277-5080					       #
# Date:   6/4/2009							       #
#									       #
################################################################################
scrname=${0##*/}	# Get shell script name
# Get shell script directory name
scrdir=${0%/$scrname}
if [ "${scrdir}" = "." ]
  then scrdir="$(pwd)"
fi
if [ -z "$scrdir" ]
then
  scrdir=$(whence $scrname)
  scrdir=${scrdir%$scrname}
fi

#
# Define useful variables
#
typeset -A diskstate		# Define associative array to hold hdisk state
hdactive="active"		# String signifying that hdisk is active
hdmissing="missing"		# String signifying that hdisk is missing
lockfile=/tmp/$scrname.lock	# Lock file indicating that script is running
tmpfile=/tmp/$scrname.$$	# Temporary file in which to save command output
sysdumpnull="/dev/sysdumpnull"	# Name of null sysdump device
rootvgname="rootvg"		# Name of root volume group
hostname=$(hostname -s)		# Short host name
notifymsg="This note was built and sent automatically by $scrdir/$scrname."
users="root"			# List of email addrs to which to send status notes

################################################################################
#									       #
# Determine if script is running from a command line.			       #
#									       #
################################################################################
/usr/bin/tty >/dev/null
if [ $? = 0 ]
then tty=true	# Script is running from a command line
else tty=false	# Script is not running from a command line
fi

################################################################################
#									       #
# Confirm that no other copy of this script is running right now and touch     #
# lock file.								       #
#									       #
################################################################################
if [ -e $lockfile ]
then
  errmsg="File $lockfile already exists, suggesting that another copy of $scrname is already running!"
  if $tty
  then
    print -u2 "$scrname: $errmsg"
  else
    subject="'$hostname: $errmsg'"
    cmd="/usr/bin/mail -s $subject $users"
    echo "$notifymsg" | eval $cmd
  fi
  exit 1
else /usr/bin/touch $lockfile
fi

################################################################################
#									       #
# Get names of primary and secondary dump devices and confirm that neither one #
# is null.								       #
#									       #
################################################################################
baddump=false	# Assume dump configuration is okay
/usr/bin/sysdumpdev -l | while read tok1 tok2 rest
do
  if [ $tok2 = $sysdumpnull ]
  then
    baddump=true	# Dump space not configured properly
    errmsg="The $tok1 dump device is $sysdumpnull!"
    if $tty
    then
      print -u2 "$scrname: $errmsg"
    else
      subject="'$hostname: $errmsg'"
      cmd="/usr/bin/mail -s $subject $users"
      echo "$notifymsg" | eval $cmd
    fi
    continue
  fi
  case $tok1 in
    primary)
      dumpdev1=$tok2
      ;;
    secondary)
      dumpdev2=$tok2
      ;;
  esac
done
if $baddump
then
  /usr/bin/rm $lockfile		# Remove lock file
  exit 2			# Exit if dump configuration was not satisfactory
fi

################################################################################
#									       #
# Check each active volume group for stale physical partitions.		       #
#									       #
################################################################################
/usr/sbin/lsvg -o | while read vgname rest
do
  stalepvs=$(/usr/sbin/lsvg $vgname | /usr/bin/grep STALE  | /usr/bin/awk '{print $3}')
  if (( "$stalepvs" > 0 ))
  then
    ############################################################################
    #									       #
    # VG has stale PVs. Check PVs which are missing from VG to see if they are #
    # now accessible.							       #
    #									       #
    ############################################################################
    foundmiss=false	# No missing disks found yet
    okaynow=true	# Assume all missing disks are okay now
    someokay=false	# Assume all missing disks are still not okay
    badstate=false	# Assume all hdisk state indicators will be recognized
    /usr/sbin/lsvg -p $vgname | tail +3 | while read hdisk state rest
    do
      ##########################################################################
      #									       #
      # Confirm that hdisk state is something expected.			       #
      #									       #
      ##########################################################################
      case $state in
        $hdmissing)
          ;;
        $hdactive)
          ;;
        *)
          badstate=true		# Don't proceed very far
          errmsg="PV $hdisk in VG $vgname is in unrecognized state '$state'
(not '$hdmissing' or '$hdactive')!  Use 'lsvg -p $vgname' to display state of PVs."
          if $tty
          then
            print -u2 "$scrname: $errmsg"
          else
            subject="'$hostname: $vgname: Unrecognizable hdisk state encountered!'"
            cmd="/usr/bin/mail -s $subject $users"
            ( echo "$errmsg" && echo "\n$notifymsg" ) | eval $cmd
          fi
      esac
      diskstate[$hdisk]=$state	# Save hdisk state for later
      ##########################################################################
      #									       #
      # If an hdisk is missing, see if first 4K bytes can be read.	       #
      #									       #
      ##########################################################################
      if [ "$state" = $hdmissing ]
      then
        foundmiss=true	# At least one hdisk is missing
        /usr/bin/dd if=/dev/r$hdisk of=/dev/null bs=4k count=1 >/dev/null 2>&1
        if [ $? = 0 ]
        then someokay=true	# At least one hdisk can be read
        else okaynow=false	# At least one hdisk can not be read
        fi
      fi
    done
    if $badstate
    then
      /usr/bin/rm $lockfile	# Remove lock file
      exit 3			# Exit if disk state was not recognizable
    fi
    if $foundmiss && ! $okaynow
    then
      if $someokay
      then
        ########################################################################
        #								       #
        # We are able to read from some, but not all, missing hdisks.          #
        #								       #
        ########################################################################
        errmsg="Some missing physical volumes in VG $vgname can be accessed, but others can not.
Use 'lsvg -p $vgname' to display physical volume states."
        if $tty
        then
          print -u2 "$scrname: $errmsg"
        else
          subject="'$hostname: $vgname: Some physical volumes still missing!'"
          cmd="/usr/bin/mail -s $subject $users"
          ( echo "$errmsg" && echo $notifymsg ) | eval $cmd
        fi
      else
        if $tty
        then
          ######################################################################
          #								       #
          # We can not read from any missing hdisks and there is a user we can #
          # inform.							       #
          #								       #
          ######################################################################
          print -u2 "$scrname: All missing volumes in VG $vgname are still inaccessible!"
        fi
      fi
    fi
    if $foundmiss && $okaynow && [ $vgname = $rootvgname ]
    then
      ##########################################################################
      #									       #
      # At least one hdisk was found missing, but we are able to read from all #
      # missing hdisks, so try to get missing hdisks back to active state.     #
      # And this is rootvg, so first determine how to deconfigure dump space.  #
      #									       #
      ##########################################################################
      ##########################################################################
      #									       #
      # Check to see if dump devices are on physical volumes which are	       #
      # missing.							       #
      #									       #
      ##########################################################################
      for lvname in ${dumpdev1#/dev/} ${dumpdev2#/dev/}
      do
        /usr/sbin/lslv -l $lvname | tail +3 | while read hdisk rest
        do
          if [ ${diskstate[$hdisk]} = $hdmissing ]
          then
            if [ $lvname = ${dumpdev1#/dev/} ]
            then pridisk=$hdisk
            else secdisk=$hdisk
            fi
          fi
        done
      done
      ##########################################################################
      #									       #
      # If both primary and secondary dump devices have missing physical       #
      # volumes, generate a message and give up.			       #
      #									       #
      ##########################################################################
      if [ ! -z "$pridisk" ] && [ ! -z "$secdisk" ]
      then
        okaynow=false	# Don't proceed further
        errmsg1="Dump device ${dumpdev1#/dev/} resides on $pridisk, which is missing!"
        errmsg2="Dump device ${dumpdev2#/dev/} resides on $secdisk, which is missing!"
        if $tty
        then
          print -u2 "$scrname: $vgname: $errmsg1"
          print -u2 "$scrname: $vgname: $errmsg2"
        else
          subject="'$hostname: $vgname: Both dump devices are on missing hdisks!'"
          cmd="/usr/bin/mail -s $subject $users"
          ( echo "$errmsg1" && echo "$errmsg2" && echo "\n$notifymsg" ) | eval $cmd
        fi
      fi
      ##########################################################################
      #									       #
      # Set parameters for sysdumpdev appropriately.			       #
      #									       #
      ##########################################################################
      if [ ! -z "$pridisk" ]
      then
        sysdumpflag="-p"
        sysdumplv=$dumpdev1
      fi
      if [ ! -z "$secdisk" ]
      then
        sysdumpflag="-s"
        sysdumplv=$dumpdev2
      fi
    fi
    if $foundmiss && $okaynow
    then
      ##########################################################################
      #									       #
      # At least one hdisk was found missing, but we are able to read from all #
      # missing hdisks, so try to get missing hdisks back to active state.     #
      #									       #
      ##########################################################################
      ##########################################################################
      #									       #
      # If appropriate, temporarily deconfigure dump space.		       #
      #									       #
      ##########################################################################
      if [ ! -z "$sysdumpflag" ]
      then /usr/bin/sysdumpdev $sysdumpflag $sysdumpnull >/dev/null
      fi
      ##########################################################################
      #									       #
      # Vary on the volume group to make missing hdisks active.		       #
      #									       #
      # Use -n flag on varyonvg so it does not launch a syncvg in the	       #
      # background.							       #
      #									       #
      ##########################################################################
      cmd="/usr/sbin/varyonvg -n $vgname"
      errmsg="Command '$cmd' failed with rc=\$varyonrc"	# Used only when failure occurs
      if $tty
      then
        print "$scrname: Varying on $vgname ..."
        $cmd			# varyonvg
        varyonrc=$?
        if [ $varyonrc = 0 ]
        then print "$scrname: VG $vgname varied on with rc=$varyonrc"
        else
          eval print -u2 \"$scrname: $errmsg!\"	# eval required to resolve $varyonrc in $errmsg
          okaynow=false		# Vary on failed, so don't attempt syncvg
        fi
      else
        $cmd >$tmpfile 2>&1		# varyonvg
        varyonrc=$?
        if [ $varyonrc != 0 ]
        then
          eval errnote=\"$errmsg:\"			# eval required to resolve $varyonrc in $errmsg
          subject="'$hostname: $vgname: All PVs in $vgname are accessible, but varyonvg failed!'"
          cmd="/usr/bin/mail -s $subject $users"
          ( echo $errnote && cat $tmpfile && echo "\n\n\n$notifymsg" ) | eval $cmd
          okaynow=false		# Vary on failed, so don't attempt syncvg
        fi
        /usr/bin/rm $tmpfile
      fi
      ##########################################################################
      #									       #
      # If appropriate, reconfigure dump space.				       #
      #									       #
      ##########################################################################
      if [ ! -z "$sysdumpflag" ]
      then
        /usr/bin/sysdumpdev $sysdumpflag $sysdumplv >/dev/null
        unset sysdumpflag	# Don't use for any VGs except rootvg
      fi
    fi
    if ! $foundmiss || $okaynow
    then
      ##########################################################################
      #									       #
      # Either no hdisks were missing or varyonvg completed successfully, so   #
      # try to resync volume group.					       #
      #									       #
      ##########################################################################
      nomissmsg="Volume group $vgname has stale physical volumes,
but there are no physical volumes in missing state."
      cmd="/usr/sbin/syncvg -v $vgname"
      errmsg="Command '$cmd' failed with rc=\$syncrc"
      if $tty
      then
        if ! $foundmiss
        then print "$scrname: $nomissmsg"
        fi
        print "$scrname: Syncing $vgname ..."
        $cmd			# syncvg
        syncrc=$?
        if [ $syncrc = 0 ]
        then print "$scrname: VG $vgname synced with rc=$syncrc"
        else eval print -u2 \"$scrname: $errmsg!\"	# eval required to resolve $syncrc in $errmsg
        fi
      else
        $cmd >$tmpfile 2>&1		# syncvg
        syncrc=$?
        if $foundmiss
        then eval errnote=\"$errmsg:\"		# eval required to resolve $syncrc in $errmsg
        else eval errnote=\"$nomissmsg\\n$errmsg:\"	# eval required to resolve $syncrc in $errmsg
        fi
        if [ $syncrc = 0 ]
        then
          subject="'$hostname: Volume group $vgname resynced'"
          cmd="/usr/bin/mail -s $subject $users"
          echo $notifymsg | eval $cmd
        else
          subject="'$hostname: Resync failed for volume group $vgname!'"
          cmd="/usr/bin/mail -s $subject $users"
          ( echo $errnote && cat $tmpfile && echo "\n\n\n$notifymsg" ) | eval $cmd
        fi
        /usr/bin/rm $tmpfile
      fi
    fi
  fi
done
/usr/bin/rm $lockfile		# Remove lock file
